import { BatchWriteItemCommand, QueryCommand, } from "@aws-sdk/client-dynamodb";
import { DeleteObjectsCommand, GetObjectCommand, ListObjectsV2Command, PutObjectCommand, } from "@aws-sdk/client-s3";
import path from "path";
import { MAX_DYNAMO_BATCH_WRITE_ITEM_COUNT } from "./constants.js";
import { debug, error, warn } from "./logger.js";
import { chunk } from "./util.js";
/** Beginning single backslash is intentional, to look for the dot + the extension. Do not escape it again. */
const CACHE_EXTENSION_REGEX = /\.(cache|fetch)$/;
export function hasCacheExtension(key) {
    return CACHE_EXTENSION_REGEX.test(key);
}
// Expected environment variables
const { CACHE_BUCKET_NAME, CACHE_BUCKET_KEY_PREFIX, CACHE_DYNAMO_TABLE, NEXT_BUILD_ID, } = process.env;
export default class S3Cache {
    client;
    dynamoClient;
    buildId;
    constructor(_ctx) {
        this.client = globalThis.S3Client;
        this.dynamoClient = globalThis.dynamoClient;
        this.buildId = NEXT_BUILD_ID;
    }
    async get(key, options) {
        if (globalThis.disableIncrementalCache) {
            return null;
        }
        const isFetchCache = typeof options === "object" ? options.fetchCache : options;
        return isFetchCache
            ? this.getFetchCache(key)
            : this.getIncrementalCache(key);
    }
    async getFetchCache(key) {
        debug("get fetch cache", { key });
        try {
            const { Body, LastModified } = await this.getS3Object(key, "fetch");
            const lastModified = await this.getHasRevalidatedTags(key, LastModified?.getTime());
            if (lastModified === -1) {
                // If some tags are stale we need to force revalidation
                return null;
            }
            if (Body === null)
                return null;
            return {
                lastModified,
                value: JSON.parse((await Body?.transformToString()) ?? "{}"),
            };
        }
        catch (e) {
            error("Failed to get fetch cache", e);
            return null;
        }
    }
    async getIncrementalCache(key) {
        try {
            const { Body, LastModified } = await this.getS3Object(key, "cache");
            const cacheData = JSON.parse((await Body?.transformToString()) ?? "{}");
            const meta = cacheData.meta;
            const lastModified = await this.getHasRevalidatedTags(key, LastModified?.getTime());
            if (lastModified === -1) {
                // If some tags are stale we need to force revalidation
                return null;
            }
            globalThis.lastModified = lastModified;
            if (cacheData.type === "route") {
                return {
                    lastModified: LastModified?.getTime(),
                    value: {
                        kind: "ROUTE",
                        body: Buffer.from(cacheData.body ?? Buffer.alloc(0)),
                        status: meta?.status,
                        headers: meta?.headers,
                    },
                };
            }
            else if (cacheData.type === "page" || cacheData.type === "app") {
                return {
                    lastModified: LastModified?.getTime(),
                    value: {
                        kind: "PAGE",
                        html: cacheData.html,
                        pageData: cacheData.type === "page" ? cacheData.json : cacheData.rsc,
                        status: meta?.status,
                        headers: meta?.headers,
                    },
                };
            }
            else if (cacheData.type === "redirect") {
                return {
                    lastModified: LastModified?.getTime(),
                    value: {
                        kind: "REDIRECT",
                        props: cacheData.props,
                    },
                };
            }
            else {
                error("Unknown cache type", cacheData);
                return null;
            }
        }
        catch (e) {
            error("Failed to get body cache", e);
            return null;
        }
    }
    async set(key, data) {
        if (globalThis.disableIncrementalCache) {
            return;
        }
        if (data?.kind === "ROUTE") {
            const { body, status, headers } = data;
            this.putS3Object(key, "cache", JSON.stringify({
                type: "route",
                body: body.toString("utf8"),
                meta: {
                    status,
                    headers,
                },
            }));
        }
        else if (data?.kind === "PAGE") {
            const { html, pageData } = data;
            const isAppPath = typeof pageData === "string";
            this.putS3Object(key, "cache", JSON.stringify({
                type: isAppPath ? "app" : "page",
                html,
                rsc: isAppPath ? pageData : undefined,
                json: isAppPath ? undefined : pageData,
                meta: { status: data.status, headers: data.headers },
            }));
        }
        else if (data?.kind === "FETCH") {
            await this.putS3Object(key, "fetch", JSON.stringify(data));
        }
        else if (data?.kind === "REDIRECT") {
            // // delete potential page data if we're redirecting
            await this.putS3Object(key, "cache", JSON.stringify({
                type: "redirect",
                props: data.props,
            }));
        }
        else if (data === null || data === undefined) {
            await this.deleteS3Objects(key);
        }
        // Write derivedTags to dynamodb
        // If we use an in house version of getDerivedTags in build we should use it here instead of next's one
        const derivedTags = data?.kind === "FETCH"
            ? data.data.tags ?? []
            : data?.kind === "PAGE"
                ? data.headers?.["x-next-cache-tags"]?.split(",") ?? []
                : [];
        debug("derivedTags", derivedTags);
        // Get all tags stored in dynamodb for the given key
        // If any of the derived tags are not stored in dynamodb for the given key, write them
        const storedTags = await this.getTagsByPath(key);
        const tagsToWrite = derivedTags.filter((tag) => !storedTags.includes(tag));
        if (tagsToWrite.length > 0) {
            await this.batchWriteDynamoItem(tagsToWrite.map((tag) => ({
                path: key,
                tag: tag,
            })));
        }
    }
    async revalidateTag(tag) {
        if (globalThis.disableDynamoDBCache || globalThis.disableIncrementalCache) {
            return;
        }
        debug("revalidateTag", tag);
        // Find all keys with the given tag
        const paths = await this.getByTag(tag);
        debug("Items", paths);
        // Update all keys with the given tag with revalidatedAt set to now
        await this.batchWriteDynamoItem(paths?.map((path) => ({
            path: path,
            tag: tag,
        })) ?? []);
    }
    // DynamoDB handling
    async getTagsByPath(path) {
        try {
            if (disableDynamoDBCache)
                return [];
            const result = await this.dynamoClient.send(new QueryCommand({
                TableName: CACHE_DYNAMO_TABLE,
                IndexName: "revalidate",
                KeyConditionExpression: "#key = :key",
                ExpressionAttributeNames: {
                    "#key": "path",
                },
                ExpressionAttributeValues: {
                    ":key": { S: this.buildDynamoKey(path) },
                },
            }));
            const tags = result.Items?.map((item) => item.tag.S ?? "") ?? [];
            debug("tags for path", path, tags);
            return tags;
        }
        catch (e) {
            error("Failed to get tags by path", e);
            return [];
        }
    }
    //TODO: Figure out a better name for this function since it returns the lastModified
    async getHasRevalidatedTags(key, lastModified) {
        try {
            if (disableDynamoDBCache)
                return lastModified ?? Date.now();
            const result = await this.dynamoClient.send(new QueryCommand({
                TableName: CACHE_DYNAMO_TABLE,
                IndexName: "revalidate",
                KeyConditionExpression: "#key = :key AND #revalidatedAt > :lastModified",
                ExpressionAttributeNames: {
                    "#key": "path",
                    "#revalidatedAt": "revalidatedAt",
                },
                ExpressionAttributeValues: {
                    ":key": { S: this.buildDynamoKey(key) },
                    ":lastModified": { N: String(lastModified ?? 0) },
                },
            }));
            const revalidatedTags = result.Items ?? [];
            debug("revalidatedTags", revalidatedTags);
            // If we have revalidated tags we return -1 to force revalidation
            return revalidatedTags.length > 0 ? -1 : lastModified ?? Date.now();
        }
        catch (e) {
            error("Failed to get revalidated tags", e);
            return lastModified ?? Date.now();
        }
    }
    async getByTag(tag) {
        try {
            if (disableDynamoDBCache)
                return [];
            const { Items } = await this.dynamoClient.send(new QueryCommand({
                TableName: CACHE_DYNAMO_TABLE,
                KeyConditionExpression: "#tag = :tag",
                ExpressionAttributeNames: {
                    "#tag": "tag",
                },
                ExpressionAttributeValues: {
                    ":tag": { S: this.buildDynamoKey(tag) },
                },
            }));
            return (
            // We need to remove the buildId from the path
            Items?.map(({ path: { S: key } }) => key?.replace(`${this.buildId}/`, "") ?? "") ?? []);
        }
        catch (e) {
            error("Failed to get by tag", e);
            return [];
        }
    }
    async batchWriteDynamoItem(req) {
        try {
            if (disableDynamoDBCache)
                return;
            await Promise.all(chunk(req, MAX_DYNAMO_BATCH_WRITE_ITEM_COUNT).map((Items) => {
                return this.dynamoClient.send(new BatchWriteItemCommand({
                    RequestItems: {
                        [CACHE_DYNAMO_TABLE ?? ""]: Items.map((Item) => ({
                            PutRequest: {
                                Item: {
                                    ...this.buildDynamoObject(Item.path, Item.tag),
                                },
                            },
                        })),
                    },
                }));
            }));
        }
        catch (e) {
            error("Failed to batch write dynamo item", e);
        }
    }
    buildDynamoKey(key) {
        // FIXME: We should probably use something else than path.join here
        // this could transform some fetch cache key into a valid path
        return path.posix.join(this.buildId, key);
    }
    buildDynamoObject(path, tags) {
        return {
            path: { S: this.buildDynamoKey(path) },
            tag: { S: this.buildDynamoKey(tags) },
            revalidatedAt: { N: `${Date.now()}` },
        };
    }
    // S3 handling
    buildS3Key(key, extension) {
        return path.posix.join(CACHE_BUCKET_KEY_PREFIX ?? "", extension === "fetch" ? "__fetch" : "", this.buildId, extension === "fetch" ? key : `${key}.${extension}`);
    }
    buildS3KeyPrefix(key) {
        return path.posix.join(CACHE_BUCKET_KEY_PREFIX ?? "", this.buildId, key);
    }
    async listS3Object(key) {
        const { Contents } = await this.client.send(new ListObjectsV2Command({
            Bucket: CACHE_BUCKET_NAME,
            // add a point to the key so that it only matches the key and
            // not other keys starting with the same string
            Prefix: `${this.buildS3KeyPrefix(key)}.`,
        }));
        return (Contents ?? []).map(({ Key }) => Key);
    }
    async getS3Object(key, extension) {
        try {
            const result = await this.client.send(new GetObjectCommand({
                Bucket: CACHE_BUCKET_NAME,
                Key: this.buildS3Key(key, extension),
            }));
            return result;
        }
        catch (e) {
            error("This error can usually be ignored : ", e);
            return { Body: null, LastModified: null };
        }
    }
    putS3Object(key, extension, value) {
        return this.client.send(new PutObjectCommand({
            Bucket: CACHE_BUCKET_NAME,
            Key: this.buildS3Key(key, extension),
            Body: value,
        }));
    }
    async deleteS3Objects(key) {
        try {
            const s3Keys = (await this.listS3Object(key)).filter((key) => key && hasCacheExtension(key));
            if (s3Keys.length === 0) {
                warn(`No s3 keys with a valid cache extension found for ${key}, see type CacheExtension in OpenNext for details`);
                return;
            }
            await this.client.send(new DeleteObjectsCommand({
                Bucket: CACHE_BUCKET_NAME,
                Delete: {
                    Objects: s3Keys.map((Key) => ({ Key })),
                },
            }));
        }
        catch (e) {
            error("Failed to delete cache", e);
        }
    }
}
